package com.sun.xml.internal.ws.transport.http;

import java.util.Arrays;
import com.sun.xml.internal.ws.util.HandlerAnnotationInfo;
import javax.xml.ws.soap.SOAPBinding;
import com.sun.xml.internal.ws.handler.HandlerChainsModel;
import javax.xml.ws.WebServiceException;
import com.sun.xml.internal.ws.util.xml.XmlUtil;
import org.xml.sax.EntityResolver;
import com.sun.xml.internal.ws.util.exception.LocatableWebServiceException;
import com.oracle.webservices.internal.api.databinding.DatabindingModeFeature;
import com.sun.xml.internal.ws.resources.ServerMessages;
import com.sun.xml.internal.ws.binding.WebServiceFeatureList;
import com.sun.xml.internal.ws.api.BindingID;
import javax.xml.ws.soap.MTOMFeature;
import com.sun.xml.internal.ws.api.WSBinding;
import com.oracle.webservices.internal.api.databinding.ExternalMetadataFeature;
import com.sun.xml.internal.ws.api.databinding.MetadataReader;
import com.sun.xml.internal.ws.streaming.Attributes;
import com.sun.xml.internal.ws.api.server.WSEndpoint;
import javax.xml.ws.WebServiceFeature;
import com.sun.xml.internal.ws.server.EndpointFactory;
import com.sun.xml.internal.ws.resources.WsservletMessages;
import java.util.ArrayList;
import java.net.URL;
import java.util.Iterator;
import java.io.FileInputStream;
import java.io.File;
import com.sun.istack.internal.NotNull;
import javax.xml.stream.XMLStreamReader;
import java.io.IOException;
import javax.xml.stream.XMLStreamException;
import com.sun.xml.internal.ws.server.ServerRtException;
import com.sun.xml.internal.ws.streaming.XMLStreamReaderUtil;
import com.sun.xml.internal.ws.streaming.TidyXMLStreamReader;
import com.sun.xml.internal.ws.api.streaming.XMLStreamReaderFactory;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.util.logging.Level;
import java.util.HashMap;
import java.util.HashSet;
import com.sun.xml.internal.ws.api.server.SDDocumentSource;
import java.util.Map;
import java.util.Set;
import com.sun.xml.internal.ws.api.server.Container;
import java.util.logging.Logger;
import java.util.List;
import javax.xml.namespace.QName;

public class DeploymentDescriptorParser<A>
{
    public static final String NS_RUNTIME = "http://java.sun.com/xml/ns/jax-ws/ri/runtime";
    public static final String JAXWS_WSDL_DD_DIR = "WEB-INF/wsdl";
    public static final QName QNAME_ENDPOINTS;
    public static final QName QNAME_ENDPOINT;
    public static final QName QNAME_EXT_METADA;
    public static final String ATTR_FILE = "file";
    public static final String ATTR_RESOURCE = "resource";
    public static final String ATTR_VERSION = "version";
    public static final String ATTR_NAME = "name";
    public static final String ATTR_IMPLEMENTATION = "implementation";
    public static final String ATTR_WSDL = "wsdl";
    public static final String ATTR_SERVICE = "service";
    public static final String ATTR_PORT = "port";
    public static final String ATTR_URL_PATTERN = "url-pattern";
    public static final String ATTR_ENABLE_MTOM = "enable-mtom";
    public static final String ATTR_MTOM_THRESHOLD_VALUE = "mtom-threshold-value";
    public static final String ATTR_BINDING = "binding";
    public static final String ATTR_DATABINDING = "databinding";
    public static final List<String> ATTRVALUE_SUPPORTED_VERSIONS;
    private static final Logger logger;
    private final Container container;
    private final ClassLoader classLoader;
    private final ResourceLoader loader;
    private final DeploymentDescriptorParser.AdapterFactory<A> adapterFactory;
    private final Set<String> names;
    private final Map<String, SDDocumentSource> docs;
    
    public DeploymentDescriptorParser(ClassLoader cl, ResourceLoader loader, Container container, DeploymentDescriptorParser.AdapterFactory<A> adapterFactory) throws MalformedURLException {
        this.names = new HashSet();
        this.docs = new HashMap();
        this.classLoader = cl;
        this.loader = loader;
        this.container = container;
        this.adapterFactory = adapterFactory;
        this.collectDocs("/WEB-INF/wsdl/");
        DeploymentDescriptorParser.logger.log(Level.FINE, "war metadata={0}", this.docs);
    }
    
    @NotNull
    public List<A> parse(String systemId, InputStream is) {
        XMLStreamReader reader = null;
        try {
            reader = new TidyXMLStreamReader(XMLStreamReaderFactory.create(systemId, is, true), is);
            XMLStreamReaderUtil.nextElementContent(reader);
            return this.parseAdapters(reader);
        }
        finally {
            if (reader != null) {
                try {
                    reader.close();
                }
                catch (XMLStreamException e) {
                    throw new ServerRtException("runtime.parser.xmlReader", new Object[] { e });
                }
            }
            try {
                is.close();
            }
            catch (IOException ex) {}
        }
    }
    
    @NotNull
    public List<A> parse(File f) throws IOException {
        FileInputStream in = new FileInputStream(f);
        try {
            return this.parse(f.getPath(), in);
        }
        finally {
            in.close();
        }
    }
    
    private void collectDocs(String dirPath) throws MalformedURLException {
        Set<String> paths = this.loader.getResourcePaths(dirPath);
        if (paths != null) {
            for (String path : paths) {
                if (path.endsWith("/")) {
                    if (path.endsWith("/CVS/")) {
                        continue;
                    }
                    if (path.endsWith("/.svn/")) {
                        continue;
                    }
                    this.collectDocs(path);
                }
                else {
                    URL res = this.loader.getResource(path);
                    this.docs.put(res.toString(), SDDocumentSource.create(res));
                }
            }
        }
    }
    
    private List<A> parseAdapters(XMLStreamReader reader) {
        if (!reader.getName().equals(DeploymentDescriptorParser.QNAME_ENDPOINTS)) {
            failWithFullName("runtime.parser.invalidElement", reader);
        }
        List<A> adapters = new ArrayList();
        Attributes attrs = XMLStreamReaderUtil.getAttributes(reader);
        String version = this.getMandatoryNonEmptyAttribute(reader, attrs, "version");
        if (!DeploymentDescriptorParser.ATTRVALUE_SUPPORTED_VERSIONS.contains(version)) {
            failWithLocalName("runtime.parser.invalidVersionNumber", reader, version);
        }
        while (XMLStreamReaderUtil.nextElementContent(reader) != 2) {
            if (reader.getName().equals(DeploymentDescriptorParser.QNAME_ENDPOINT)) {
                attrs = XMLStreamReaderUtil.getAttributes(reader);
                String name = this.getMandatoryNonEmptyAttribute(reader, attrs, "name");
                if (!this.names.add(name)) {
                    DeploymentDescriptorParser.logger.warning(WsservletMessages.SERVLET_WARNING_DUPLICATE_ENDPOINT_NAME());
                }
                String implementationName = this.getMandatoryNonEmptyAttribute(reader, attrs, "implementation");
                Class<?> implementorClass = this.getImplementorClass(implementationName, reader);
                MetadataReader metadataReader = null;
                ExternalMetadataFeature externalMetadataFeature = null;
                XMLStreamReaderUtil.nextElementContent(reader);
                if (reader.getEventType() != 2) {
                    externalMetadataFeature = this.configureExternalMetadataReader(reader);
                    if (externalMetadataFeature != null) {
                        metadataReader = externalMetadataFeature.getMetadataReader(implementorClass.getClassLoader(), false);
                    }
                }
                QName serviceName = this.getQNameAttribute(attrs, "service");
                if (serviceName == null) {
                    serviceName = EndpointFactory.getDefaultServiceName(implementorClass, metadataReader);
                }
                QName portName = this.getQNameAttribute(attrs, "port");
                if (portName == null) {
                    portName = EndpointFactory.getDefaultPortName(serviceName, implementorClass, metadataReader);
                }
                String enable_mtom = this.getAttribute(attrs, "enable-mtom");
                String mtomThreshold = this.getAttribute(attrs, "mtom-threshold-value");
                String dbMode = this.getAttribute(attrs, "databinding");
                String bindingId = this.getAttribute(attrs, "binding");
                if (bindingId != null) {
                    bindingId = getBindingIdForToken(bindingId);
                }
                WSBinding binding = createBinding(bindingId, implementorClass, enable_mtom, mtomThreshold, dbMode);
                if (externalMetadataFeature != null) {
                    binding.getFeatures().mergeFeatures(new WebServiceFeature[] { externalMetadataFeature }, true);
                }
                String urlPattern = this.getMandatoryNonEmptyAttribute(reader, attrs, "url-pattern");
                boolean handlersSetInDD = this.setHandlersAndRoles(binding, reader, serviceName, portName);
                EndpointFactory.verifyImplementorClass(implementorClass, metadataReader);
                SDDocumentSource primaryWSDL = this.getPrimaryWSDL(reader, attrs, implementorClass, metadataReader);
                WSEndpoint<?> endpoint = WSEndpoint.create(implementorClass, !handlersSetInDD, null, serviceName, portName, this.container, binding, primaryWSDL, this.docs.values(), this.createEntityResolver(), false);
                adapters.add(this.adapterFactory.createAdapter(name, urlPattern, endpoint));
            }
            else {
                failWithLocalName("runtime.parser.invalidElement", reader);
            }
        }
        return adapters;
    }
    
    private static WSBinding createBinding(String ddBindingId, Class implClass, String mtomEnabled, String mtomThreshold, String dataBindingMode) {
        MTOMFeature mtomfeature = null;
        if (mtomEnabled != null) {
            if (mtomThreshold != null) {
                mtomfeature = new MTOMFeature((boolean)Boolean.valueOf(mtomEnabled), (int)Integer.valueOf(mtomThreshold));
            }
            else {
                mtomfeature = new MTOMFeature((boolean)Boolean.valueOf(mtomEnabled));
            }
        }
        BindingID bindingID;
        WebServiceFeatureList features;
        if (ddBindingId != null) {
            bindingID = BindingID.parse(ddBindingId);
            features = bindingID.createBuiltinFeatureList();
            if (checkMtomConflict((MTOMFeature)features.get(MTOMFeature.class), mtomfeature)) {
                throw new ServerRtException(ServerMessages.DD_MTOM_CONFLICT(ddBindingId, mtomEnabled), new Object[0]);
            }
        }
        else {
            bindingID = BindingID.parse(implClass);
            features = new WebServiceFeatureList();
            if (mtomfeature != null) {
                features.add(mtomfeature);
            }
            features.addAll(bindingID.createBuiltinFeatureList());
        }
        if (dataBindingMode != null) {
            features.add(new DatabindingModeFeature(dataBindingMode));
        }
        return bindingID.createBinding(features.toArray());
    }
    
    private static boolean checkMtomConflict(MTOMFeature lhs, MTOMFeature rhs) {
        return lhs != null && rhs != null && (lhs.isEnabled() ^ rhs.isEnabled());
    }
    
    @NotNull
    public static String getBindingIdForToken(@NotNull String lexical) {
        if (lexical.equals("##SOAP11_HTTP")) {
            return "http://schemas.xmlsoap.org/wsdl/soap/http";
        }
        if (lexical.equals("##SOAP11_HTTP_MTOM")) {
            return "http://schemas.xmlsoap.org/wsdl/soap/http?mtom=true";
        }
        if (lexical.equals("##SOAP12_HTTP")) {
            return "http://www.w3.org/2003/05/soap/bindings/HTTP/";
        }
        if (lexical.equals("##SOAP12_HTTP_MTOM")) {
            return "http://www.w3.org/2003/05/soap/bindings/HTTP/?mtom=true";
        }
        if (lexical.equals("##XML_HTTP")) {
            return "http://www.w3.org/2004/08/wsdl/http";
        }
        return lexical;
    }
    
    private SDDocumentSource getPrimaryWSDL(XMLStreamReader xsr, Attributes attrs, Class<?> implementorClass, MetadataReader metadataReader) {
        String wsdlFile = this.getAttribute(attrs, "wsdl");
        if (wsdlFile == null) {
            wsdlFile = EndpointFactory.getWsdlLocation(implementorClass, metadataReader);
        }
        if (wsdlFile != null) {
            if (!wsdlFile.startsWith("WEB-INF/wsdl")) {
                DeploymentDescriptorParser.logger.log(Level.WARNING, "Ignoring wrong wsdl={0}. It should start with {1}. Going to generate and publish a new WSDL.", new Object[] { wsdlFile, "WEB-INF/wsdl" });
                return null;
            }
            URL wsdl;
            try {
                wsdl = this.loader.getResource(47 + wsdlFile);
            }
            catch (MalformedURLException e) {
                throw new LocatableWebServiceException(ServerMessages.RUNTIME_PARSER_WSDL_NOT_FOUND(wsdlFile), e, xsr);
            }
            if (wsdl == null) {
                throw new LocatableWebServiceException(ServerMessages.RUNTIME_PARSER_WSDL_NOT_FOUND(wsdlFile), xsr);
            }
            SDDocumentSource docInfo = (SDDocumentSource)this.docs.get(wsdl.toExternalForm());
            assert docInfo != null;
            return docInfo;
        }
        else {
            return null;
        }
    }
    
    private EntityResolver createEntityResolver() {
        try {
            return XmlUtil.createEntityResolver(this.loader.getCatalogFile());
        }
        catch (MalformedURLException e) {
            throw new WebServiceException(e);
        }
    }
    
    protected String getAttribute(Attributes attrs, String name) {
        String value = attrs.getValue(name);
        if (value != null) {
            value = value.trim();
            return value;
        }
        return value;
    }
    
    protected QName getQNameAttribute(Attributes attrs, String name) {
        String value = this.getAttribute(attrs, name);
        if (value == null || value.equals("")) {
            return null;
        }
        return QName.valueOf(value);
    }
    
    protected String getNonEmptyAttribute(XMLStreamReader reader, Attributes attrs, String name) {
        String value = this.getAttribute(attrs, name);
        if (value != null && value.equals("")) {
            failWithLocalName("runtime.parser.invalidAttributeValue", reader, name);
            return value;
        }
        return value;
    }
    
    protected String getMandatoryAttribute(XMLStreamReader reader, Attributes attrs, String name) {
        String value = this.getAttribute(attrs, name);
        if (value == null) {
            failWithLocalName("runtime.parser.missing.attribute", reader, name);
            return value;
        }
        return value;
    }
    
    protected String getMandatoryNonEmptyAttribute(XMLStreamReader reader, Attributes attributes, String name) {
        String value = this.getAttribute(attributes, name);
        if (value == null) {
            failWithLocalName("runtime.parser.missing.attribute", reader, name);
        }
        else if (value.equals("")) {
            failWithLocalName("runtime.parser.invalidAttributeValue", reader, name);
        }
        return value;
    }
    
    protected boolean setHandlersAndRoles(WSBinding binding, XMLStreamReader reader, QName serviceName, QName portName) {
        if (reader.getEventType() == 2 || !reader.getName().equals(HandlerChainsModel.QNAME_HANDLER_CHAINS)) {
            return false;
        }
        HandlerAnnotationInfo handlerInfo = HandlerChainsModel.parseHandlerFile(reader, this.classLoader, serviceName, portName, binding);
        binding.setHandlerChain(handlerInfo.getHandlers());
        if ((binding instanceof SOAPBinding)) {
            ((SOAPBinding)binding).setRoles(handlerInfo.getRoles());
        }
        XMLStreamReaderUtil.nextContent(reader);
        return true;
    }
    
    protected ExternalMetadataFeature configureExternalMetadataReader(XMLStreamReader reader) {
        ExternalMetadataFeature.Builder featureBuilder = null;
        while (DeploymentDescriptorParser.QNAME_EXT_METADA.equals(reader.getName())) {
            if (reader.getEventType() == 1) {
                Attributes attrs = XMLStreamReaderUtil.getAttributes(reader);
                String file = this.getAttribute(attrs, "file");
                if (file != null) {
                    if (featureBuilder == null) {
                        featureBuilder = ExternalMetadataFeature.builder();
                    }
                    featureBuilder.addFiles(new File[] { new File(file) });
                }
                String res = this.getAttribute(attrs, "resource");
                if (res != null) {
                    if (featureBuilder == null) {
                        featureBuilder = ExternalMetadataFeature.builder();
                    }
                    featureBuilder.addResources(new String[] { res });
                }
            }
            XMLStreamReaderUtil.nextElementContent(reader);
        }
        return this.buildFeature(featureBuilder);
    }
    
    private ExternalMetadataFeature buildFeature(ExternalMetadataFeature.Builder builder) {
        return (builder != null) ? builder.build() : null;
    }
    
    protected static void fail(String key, XMLStreamReader reader) {
        DeploymentDescriptorParser.logger.log(Level.SEVERE, "{0}{1}", new Object[] { key, Integer.valueOf(reader.getLocation().getLineNumber()) });
        throw new ServerRtException(key, new Object[] { Integer.toString(reader.getLocation().getLineNumber()) });
    }
    
    protected static void failWithFullName(String key, XMLStreamReader reader) {
        throw new ServerRtException(key, new Object[] { Integer.valueOf(reader.getLocation().getLineNumber()), reader.getName() });
    }
    
    protected static void failWithLocalName(String key, XMLStreamReader reader) {
        throw new ServerRtException(key, new Object[] { Integer.valueOf(reader.getLocation().getLineNumber()), reader.getLocalName() });
    }
    
    protected static void failWithLocalName(String key, XMLStreamReader reader, String arg) {
        throw new ServerRtException(key, new Object[] { Integer.valueOf(reader.getLocation().getLineNumber()), reader.getLocalName(), arg });
    }
    
    protected Class loadClass(String name) {
        try {
            return Class.forName(name, true, this.classLoader);
        }
        catch (ClassNotFoundException e) {
            DeploymentDescriptorParser.logger.log(Level.SEVERE, e.getMessage(), e);
            throw new ServerRtException("runtime.parser.classNotFound", new Object[] { name });
        }
    }
    
    private Class getImplementorClass(String name, XMLStreamReader xsr) {
        try {
            return Class.forName(name, true, this.classLoader);
        }
        catch (ClassNotFoundException e) {
            DeploymentDescriptorParser.logger.log(Level.SEVERE, e.getMessage(), e);
            throw new LocatableWebServiceException(ServerMessages.RUNTIME_PARSER_CLASS_NOT_FOUND(name), e, xsr);
        }
    }
    
    static {
        QNAME_ENDPOINTS = new QName("http://java.sun.com/xml/ns/jax-ws/ri/runtime", "endpoints");
        QNAME_ENDPOINT = new QName("http://java.sun.com/xml/ns/jax-ws/ri/runtime", "endpoint");
        QNAME_EXT_METADA = new QName("http://java.sun.com/xml/ns/jax-ws/ri/runtime", "external-metadata");
        ATTRVALUE_SUPPORTED_VERSIONS = Arrays.asList(new String[] { "2.0", "2.1" });
        logger = Logger.getLogger("com.sun.xml.internal.ws.server.http");
    }
    
    public interface AdapterFactory<A>
    {
        public abstract A createAdapter(String p0, String p1, WSEndpoint<?> p2);
    }
}
